#include <stdlib.h>

#include "kernel.h"
#include "mutex.h"

/*
--------------------------------------------------------------------------------
CreateMutex - create an initially unoccupied mutex
--------------------------------------------------------------------------------
*/

Mutex_t *
CreateMutex(char *name)
{
	Mutex_t *mut = Malloc(sizeof(Mutex_t));

	mut->sem = CreateSem(name, 1);
	return mut;
}

/*
--------------------------------------------------------------------------------
DeleteMutex - destroy a mutex
--------------------------------------------------------------------------------
*/

void 			
DeleteMutex(Mutex_t *mut)
{
	DeleteSem(mut->sem);
	Free(mut);
}

/*
--------------------------------------------------------------------------------
EnterMutex, EnterMutexCond, EnterMutexTimed - occupy a mutex

If the return status is true, the task owns the mutex. This operation may be
nested; to free the mutex, LeaveMutex() should be called once for each time
the mutex was occupied.
--------------------------------------------------------------------------------
*/

bool			
EnterMutex(Mutex_t *mut)
{
	return EnterMutexTimed(mut, FOREVER);
}

bool			
EnterMutexCond(Mutex_t *mut)
{
	return EnterMutexTimed(mut, 0);
}

bool			
EnterMutexTimed(Mutex_t *mut, Time_t msecs)
{
	if ( mut->owner == mt_curr_task )
	{
		mut->use_count++;
		return true;
	}
	if ( WaitSemTimed(mut->sem, msecs) )
	{
		mut->owner = mt_curr_task;
		mut->use_count = 1;
		return true;
	}
	return false;
}

/*
--------------------------------------------------------------------------------
LeaveMutex - leave a mutex

Should be called once for each time the mutex was occupied (i.e. for each
time EnterMutex() was called successfully).
--------------------------------------------------------------------------------
*/

void			
LeaveMutex(Mutex_t *mut)
{
	if ( mut->owner != mt_curr_task )
		Panic("LeaveMutex: task does not own the mutex");

	if ( !--mut->use_count )
	{
		mut->owner = NULL;
		SignalSem(mut->sem);
	}
}

