#include <stdio.h>
#include <conio.h>
#include "mtask.h"

bool main_sleeping;
unsigned callback_count;

struct tcontrol
{
	Time_t msecs;
	unsigned counter;
};

struct tcontrol tc1 = { 1 }, tc2 = { 10 }, tc3 = { 100 };

void
ptask(void *arg)
{
	struct tcontrol *tc = arg;

	while ( true )
	{
		tc->counter++;
		Delay(tc->msecs);
	}
}

void
better_ptask(void *arg)
{
	struct tcontrol *tc = arg;
	Time_t deadline = GetTime();

	while ( true )
	{
		tc->counter++;
		DelayUntil(deadline += tc->msecs);
	}
}

void
callback(void)
{
	if ( main_sleeping )
		callback_count++;
}

int 
main(int argc, char **argv)
{
	Time_t mspertick = 1;
	TaskFunc_t f = ptask;

	if ( argc >= 2 )
		mspertick = atoi(argv[1]);
	if ( argc >= 3 && atoi(argv[2]) )
		f = better_ptask;
	if ( argc >= 4 )
		tc1.msecs = atoi(argv[3]);
	if ( argc >= 5 )
		tc2.msecs = atoi(argv[4]);
	if ( argc >= 6 )
		tc3.msecs = atoi(argv[5]);

	printf("Timer interrupt every %u msecs\n", mspertick);
	printf("Periodic tasks use %s\n", f == ptask ? "Delay()" : "DelayUntil()");

	InitMTask(mspertick, 1000);

	// These tasks have lower priority so they will start running
	// only when the main task blocks.

	Ready(CreateTask(f, 1000, &tc1, NULL, DEFAULT_PRIO-1));
	Ready(CreateTask(f, 1000, &tc2, NULL, DEFAULT_PRIO-1));
	Ready(CreateTask(f, 1000, &tc3, NULL, DEFAULT_PRIO-1));

	// Install timer callback
	// This will count timer interrupts while the main task sleeps.

	SetTimerCallback(callback);

	// Periodic tasks start running and callback starts counting here.
	// Main task gets control back in 1 second

	puts("Main task sleeping for one second...");
	main_sleeping = true;
	Delay(1000);
	main_sleeping = false;

	// Slept 1 second

	puts("Done.\nPeriodic task cycles:");
	printf("%9u msecs: %6u (expected: %6u)\n", tc1.msecs, tc1.counter, 1000/tc1.msecs);
	printf("%9u msecs: %6u (expected: %6u)\n", tc2.msecs, tc2.counter, 1000/tc2.msecs);
	printf("%9u msecs: %6u (expected: %6u)\n", tc3.msecs, tc3.counter, 1000/tc3.msecs);
	printf("Timer callbacks: %6u (expected: %6u)\n", callback_count, 1000/mspertick);

	return 0;
}
